
class ValueObservation
{
public:
	//observation is observation_constant >= gamma if observed_above_constant is true,
	// and observation_constant >= gamma otherwise
	ValueObservation(double lower_bound, double upper_bound, double cur_time)
	{
		lowerBound = lower_bound;
		upperBound = upper_bound;
		timeStamp = cur_time;
	}

	//returns the value of the probability density function at value for the given discount factor, current time, and probability of error
	inline double GetPDFValue(double value, double discount_factor, double cur_time, double probability_of_error = 0.0)
	{
		if(probability_of_error == 0.0)
		{
			if(value >= lowerBound && value <= upperBound)
				return (1 - (1 - pow(discount_factor, cur_time - timeStamp) ) * ( 1 - (upperBound - lowerBound)) )
						/ (upperBound - lowerBound);
			else
				return 1 - pow(discount_factor, cur_time - timeStamp);
		}
		else //need to deal with error
		{
			if(value >= lowerBound && value <= upperBound)
				return 1 - (probability_of_error - (1 - (upperBound - lowerBound))) * pow(discount_factor, cur_time - timeStamp)
						/ (upperBound - lowerBound);
			else
				return 1 - (1 - probability_of_error / (1 - (upperBound - lowerBound))) * pow(discount_factor, cur_time - timeStamp);
		}
	}

	static bool CompareTimes(ValueObservation a, ValueObservation b)
	{
		return a.timeStamp > b.timeStamp;
	}

	double upperBound;
	double lowerBound;
	double timeStamp;
};

//creates a piecewise PDF function on [0,1]
class PiecewisePDF
{
public:
	PiecewisePDF() {}

	std::vector<double> ranges;
	std::vector<double> values;

	double expectedValue;	//computed and stored for convenience
};

class ObservationHistory
{
public:
	//builds an observation history for a uniform distribution between a minimum and maximum value
	ObservationHistory(double min_value = 0.0, double max_value = 1.0)
	{
		minValue = min_value;
		maxValue = max_value;
		probabilityOfError = 0.0;
	}

	ObservationHistory(const ObservationHistory &oh)
	{
		observations.reserve(oh.observations.size());
		for(int i = 0; i < oh.observations.size(); i++)
			observations.push_back(ValueObservation(oh.observations[i]));

		minValue = oh.minValue;
		maxValue = oh.maxValue;
		probabilityOfError = oh.probabilityOfError;
	}

	//adds a new observation to the history
	// observation is observation_constant >= gamma if observed_above_constant is true,
	// and observation_constant >= gamma otherwise
	void AddObservation(double lower_bound, double upper_bound, double cur_time)
	{
		observations.push_back(ValueObservation(toZeroOneRange(lower_bound), toZeroOneRange(upper_bound), cur_time));
	}

	//adds events from another observation history
	// right now, it assumes that each ObservationHistory has the same min and max values
	void CombineObservationHistories(ObservationHistory &oh)
	{
		observations.reserve(observations.size() + oh.observations.size());
		observations.insert(observations.end(), oh.observations.begin(), oh.observations.end());
	}

	//goes through the observations and removes any that conflict with newer data
	// returns the time of the most recent observation that conflicts with all data earlier than it
	double RemoveInconsistantObservations()
	{
		//keep ANDing all observations - take the lowest upperBound and highest lowerBound
		// if one doesn't fit in that range, then it's out
		double minUpperBound = 1.0;
		double maxLowerBound = 0.0;

		//start at the earliest time and work to the later times until the first inconsistancy
		std::sort(observations.begin(), observations.end(), ValueObservation::CompareTimes);
		std::vector<ValueObservation>::iterator iter;
		for(iter = observations.begin(); iter != observations.end(); iter++)
		{
			if(iter->upperBound < minUpperBound)
				minUpperBound = iter->upperBound;
			if(iter->lowerBound > maxLowerBound)
				maxLowerBound = iter->lowerBound;
			if(minUpperBound < maxLowerBound)
				break;
		}

		//if didn't find anything, return 0
		if(iter == observations.end())
			return 0.0;

		//delete all previous events before the conflicting observation
		double first_bad_observation_time = iter->timeStamp;
		observations.erase(iter, observations.end());
		return first_bad_observation_time;
	}

	//builds a PiecewisePDF from all ValueObservations for a given discount factor, time, and probability of error
	void ConstructPDF(PiecewisePDF &pdf, double discount_factor, double cur_time, double probability_of_error)
	{
		//accumulate all value ranges
		for(std::vector<ValueObservation>::iterator iter = observations.begin(); iter != observations.end(); iter++)
		{
			pdf.ranges.push_back(iter->lowerBound);
			pdf.ranges.push_back(iter->upperBound);
		}

		//add in the beginning and end of the line
		pdf.ranges.push_back(0.0);
		pdf.ranges.push_back(1.0);

		//find the unique values to get the ranges
		std::sort(pdf.ranges.begin(), pdf.ranges.end());
		std::vector<double>::iterator new_end = std::unique(pdf.ranges.begin(), pdf.ranges.end());
		pdf.ranges.erase(new_end, pdf.ranges.end());

		//accumulate the total area of probability and expected value as Riemann sums
		double accum_area = 0.0;
		double accum_value = 0.0;
		for(std::vector<double>::iterator iter = pdf.ranges.begin(); iter+1 != pdf.ranges.end(); iter++)
		{
			double ave_x = ( (*(iter+1)) + (*iter) ) / 2.0; //position in the middle of the two values
			double x_dist = ( (*(iter+1)) - (*iter) ); //distance
			double pdf_value = 1.0;

			//find the product of the PDF values
			for(std::vector<ValueObservation>::iterator obs = observations.begin(); obs != observations.end(); obs++)
			{
				pdf_value *= obs->GetPDFValue(ave_x, discount_factor, cur_time, probability_of_error);
				if(pdf_value == 0.0)	//don't need to compute any more if it's 0
					break;
			}

			//accumulate the new areas/values
			double v = pdf_value * x_dist;
			accum_area += v;
			accum_value += ave_x * v;
			pdf.values.push_back(pdf_value);	//store value at respective spot in pdf
		}

		if(accum_area == 0.0)
			return;

		//now normalize all pdf values
		for(std::vector<double>::iterator iter = pdf.values.begin(); iter != pdf.values.end(); iter++)
			(*iter) /= accum_area;

		pdf.expectedValue = accum_value / accum_area;
	}

	//returns the expected discount factor taking into account all of the observations discounted through time,
	double GetExpectedValue(double discount_factor, double cur_time)
	{
		PiecewisePDF pdf;
		ConstructPDF(pdf, discount_factor, cur_time, probabilityOfError);
		return toMinMaxRange(pdf.expectedValue);
	}

	//computes the relative entropy of the new observation history oh over the current distribution
	// note that cur_time MUST be > than the time of all the events so that no probability is 0, otherwise
	// relative entropy doesn't work
	double GetRelativeEntropyGivenNewObservationHistory(ObservationHistory &oh, double discount_factor, double cur_time, double probability_of_error = 0.0)
	{
		PiecewisePDF pdf, pdfoh;
		ConstructPDF(pdf, discount_factor, cur_time, probability_of_error);
		oh.ConstructPDF(pdfoh, discount_factor, cur_time, probability_of_error);

		std::vector<double>::iterator pdf_range_iter = pdf.ranges.begin()+1; //skip the 0
		std::vector<double>::iterator pdf_value_iter = pdf.values.begin();
		std::vector<double>::iterator pdf_oh_range_iter = pdfoh.ranges.begin()+1; //skip the 0
		std::vector<double>::iterator pdf_oh_value_iter = pdfoh.values.begin();

		//whlie have something left in either PDF
		double cur_min_range = 0.0;	//where to start
		double accum_entropy = 0.0; //accumulate total entropy
		for(;;)
		{
			//get up to next step
			double max_range = std::min(*pdf_range_iter, *pdf_oh_range_iter);
			double p = *pdf_value_iter;
			double q = *pdf_oh_value_iter;

			double f = 0.0;
			if(p != 0 && q != 0.0)
				f = (max_range - cur_min_range) * (p * log(p / q) / log(2.0));
			accum_entropy += f;

			if((*pdf_range_iter) <= (*pdf_oh_range_iter))
			{
				//advance pdf
				cur_min_range = (*pdf_range_iter);
				pdf_range_iter++;
				pdf_value_iter++;
				//see if done with the list
				if(pdf_range_iter == pdf.ranges.end())
					break;
			}
			else //(*pdf_range_iter) > (*pdf_oh_range_iter)
			{
				//advance pdf_oh
				cur_min_range = (*pdf_oh_range_iter);
				pdf_oh_range_iter++;
				pdf_oh_value_iter++;
				//see if done with the list
				if(pdf_oh_range_iter == pdfoh.ranges.end())
					break;
			}

		}
		return accum_entropy;
	}

	//evaluates the differential entropy relative to the uniform distribution
	// used by the function GetProbabilityOfErrorWithMaximumEntropy
	class ObservationDifferentialEntropyEvaluator : public Evaluator
	{
	public:
		ObservationDifferentialEntropyEvaluator(ObservationHistory *oh, double discount_factor, double cur_time)
		{	observationHistory = oh;	discountFactor = discount_factor;	curTime = cur_time;		}

		double Evaluate(std::vector<double> v)
		{
			double p = v[0];
			ObservationHistory noObservations(observationHistory->minValue, observationHistory->maxValue);
			return noObservations.GetRelativeEntropyGivenNewObservationHistory(*observationHistory, discountFactor, curTime, p);
		}
		
		ObservationHistory *observationHistory;
		double discountFactor;
		double curTime;
	};

	//returns the error probability that maximizes the differential entropy relative to the uniform distribution
	double GetProbabilityOfErrorWithMaximumEntropy(double discount_factor, double cur_time)
	{
		ObservationDifferentialEntropyEvaluator odiffent(this, discount_factor, cur_time);
		//only need to check probabilities between 0 and .5, because if > .5, then better off taking the opposite signal
		// (and there may be a value of the error probability on the range of [.5,1] that will give an equal entropy value)
		return MaximizeEvaluator(odiffent, 0.0, 0.5);
	}

	//***need this function???
	double GetProbabilityOfNewValueObservation(ValueObservation &vo)
	{
		return 0.0;
	}

	//deals with inconsistant observations.  if perfect_signalling is true, then it removes
	// old observations that conflict because the agent likely underwent replacement.
	// if perfect_signalling is false, then it recomputes probabilityOfError based on the observations
	void HandleInconsistantObservations(double discount_factor, double cur_time, bool perfect_signalling)
	{
		if(perfect_signalling)
			RemoveInconsistantObservations();
//		else
//			probabilityOfError = GetProbabilityOfErrorWithMaximumEntropy(discount_factor, cur_time);
	}

	// v in [minValue, maxValue] -> [0,1]
	double toZeroOneRange(double v)
	{
		return (v - minValue) / (maxValue - minValue);
	}

	// v in [0,1] -> [minValue, maxValue]
	double toMinMaxRange(double v)
	{
		return minValue + v * (maxValue - minValue);
	}

	std::vector<ValueObservation> observations;
	//allowed range of observations
	double minValue, maxValue;

	//probability that any given observation is erroneous
	double probabilityOfError;
};

void DeepCopyObservationHistory(std::vector<ObservationHistory> &src, std::vector<ObservationHistory> &dest)
{
	dest.clear();
	dest.reserve(src.size());
	for(int i = 0; i < src.size(); i++)
		dest.push_back(ObservationHistory(src[i]));
}
