#include "MersenneTwister.h"
#include "statsmath.h"


class RandomVariable
{
public:
	virtual double GetNextRand() = 0;

	virtual double GetMinimumValue() = 0;
	virtual double GetMaximumValue() = 0;

	//returns the value of the PDF at the position specified
	virtual double GetPDFValue(double x) = 0;

	//returns the probability [0,1] that a random number will be in the specified range
	virtual double GetProbabilityInRange(double low, double high) = 0;

	//returns the expected value
	virtual double GetExpectedValue() = 0;

	//returns the expected value of the random number, given that that the random number is in the specified range
	virtual double GetExpectedValueInRange(double low, double high) = 0;
};


class UniformDistribution : public RandomVariable
{
public:
	UniformDistribution(double lower_bound, double upper_bound, bool closed_upper_bound = true)
	{
		upperBound = upper_bound;
		lowerBound = lower_bound;
		closedUpperBound = closed_upper_bound;
	}

	double GetNextRand()
	{
		if(closedUpperBound)
			return lowerBound + genrand_real1() * (upperBound - lowerBound);
		else
			return lowerBound + genrand_real2() * (upperBound - lowerBound);
	}

	double GetMinimumValue()
	{	return lowerBound;	}

	double GetMaximumValue()
	{	return upperBound;	}

	double GetPDFValue(double x)
	{
		return 1.0 / (upperBound - lowerBound);
	}

	double GetProbabilityInRange(double low, double high)
	{
		//cap to range of variable
		low = ClampVariableInBounds(low);
		high = ClampVariableInBounds(high);

		//translate to [0,1]
		low = (low - lowerBound) / (upperBound - lowerBound);
		high = (high - lowerBound) / (upperBound - lowerBound);

		return high - low;
	}

	double GetExpectedValue()
	{
		return (upperBound + lowerBound) / 2;
	}

	double GetExpectedValueInRange(double low, double high)
	{
		//cap to range of variable
		low = ClampVariableInBounds(low);
		high = ClampVariableInBounds(high);

		//translate to [0,1]
		low = (low - lowerBound) / (upperBound - lowerBound);
		high = (high - lowerBound) / (upperBound - lowerBound);

		//take average position and translate back to [lowerBound,upperBound]
		return ((high + low) / 2) * (upperBound - lowerBound) + lowerBound;
	}

	double ClampVariableInBounds(double v)
	{
		return std::max(lowerBound, std::min(upperBound, v));
	}

	double upperBound, lowerBound;
	bool closedUpperBound;
};

class StratifiedDiscreteUniformDistribution : public RandomVariable
{
public:
	StratifiedDiscreteUniformDistribution(int lower_bound, int upper_bound, int num_samples)
	{
		upperBound = upper_bound;
		lowerBound = lower_bound;

		//make proper amount of numbers in order
		samples.reserve(num_samples);
		for(int i = 0; i < num_samples; i++)
			samples.push_back(i % (upper_bound - lower_bound + 1) + lower_bound);

		//shuffle numbers
		RandomVariable *random_sample = new UniformDistribution(0, 1);
		int i = num_samples;
		while(--i > 0) 
		{
			int k = (int)(random_sample->GetNextRand() * (i + 1));  // [0,i]
			int tmp = samples[i];
			samples[i] = samples[k];
			samples[k] = tmp;
		}

		delete random_sample;
	}

	double GetNextRand()
	{
		double v = samples.back();
		samples.pop_back();
		return v;
	}

	double GetMinimumValue()
	{	return lowerBound;	}

	double GetMaximumValue()
	{	return upperBound;	}

	double GetPDFValue(double x)
	{
		return 1.0 / (upperBound - lowerBound);
	}

	double GetProbabilityInRange(double low, double high)
	{
		//cap to range of variable
		low = ClampVariableInBounds(low);
		high = ClampVariableInBounds(high);

		//translate to [0,1]
		low = (low - lowerBound) / (upperBound - lowerBound);
		high = (high - lowerBound) / (upperBound - lowerBound);

		return high - low;
	}

	double GetExpectedValue()
	{
		return (upperBound + lowerBound) / 2;
	}

	double GetExpectedValueInRange(double low, double high)
	{
		//cap to range of variable
		low = ClampVariableInBounds(low);
		high = ClampVariableInBounds(high);

		//translate to [0,1]
		low = (low - lowerBound) / (upperBound - lowerBound);
		high = (high - lowerBound) / (upperBound - lowerBound);

		//take average position and translate back to [lowerBound,upperBound]
		return ((high + low) / 2) * (upperBound - lowerBound) + lowerBound;
	}

	double ClampVariableInBounds(double v)
	{
		return std::max((double)lowerBound, std::min((double)upperBound, v));
	}

	int upperBound, lowerBound;
	std::vector<int> samples;
};


class ExponentialDistribution : public RandomVariable
{
public:
	ExponentialDistribution(double mean)
	{
		lambda = 1/mean;
	}

	double GetNextRand()
	{
		return -1/lambda * log(genrand_real1());
	}

	double GetMinimumValue()
	{	return 0.0;	}

	double GetMaximumValue()
	{	
		//return INFINITY;
		//***can't return infinity, because won't work with solver :(
		return 100000;
	}

	double GetPDFValue(double x)
	{
		return lambda * exp(-lambda * x);
	}

	double GetProbabilityInRange(double low, double high)
	{
		//cap to range of variable
		low = std::max(low, 0.0);
		if(high == INFINITY)
			return exp(-lambda * low);
		return (1 - exp(-lambda * high)) - (1 - exp(-lambda * low));
	}

	double GetExpectedValue()
	{
		return 1/lambda;
	}

	double GetExpectedValueInRange(double low, double high)
	{
		//cap to range of variable
		low = std::max(low, 0.0);
		
		if(high == INFINITY)
			return exp(-lambda*low)*(lambda*low + 1)/lambda;
		return exp(-lambda*low)*(lambda*low + 1)/lambda - exp(-lambda*high)*(lambda*high + 1)/lambda;
	}

	double lambda;
};
