
//this should be extended for the actual evaluation functions
class Evaluator
{
public:
	virtual double Evaluate(std::vector<double>) { return -1; };
};

template <class T>
T sq(T n)
{	return n * n;	}

const double epsilon2 = sq(0.015625);

//returns 3^(n+1)
inline int nth_pow_of_3(int n)
{
	int d = 1;
	for(int i = 0; i < n; i++) d *= 3;
	return d;
}

//takes in a vector of Evaluator pointers to use as functions to evaluate the inputs,
// as well as the vectors min_input and max_input for the bounds on each variable (each element position is a different variable)
// tolerance is the amount of precision required for each variable - the function will keep subdividing until the smallest tolerance is reached
// returns the resulting set of variables in results computed by using using a multivariate bisection method.
// Note that min_input, max_input, and results must all be of the same length
// The parameter minimum_descent_level may be used to force the algorithm to subdivide each space a minimum number of times
// to help find roots of functions that are not bijective
bool SolveSimultaneousRoots(std::vector<Evaluator *> &sre,
							std::vector<double> &min_input, std::vector<double> &max_input, std::vector<double> &tolerance_input,
							std::vector<double> &results,
							int minimum_descent_level = 0)
{
	int input_dimension = min_input.size();
	int num_evaluators = sre.size();

	//we define a corner as a vertex of a region of the inputs
	int total_num_corners = nth_pow_of_3(input_dimension);	//3^n corners to deal with... low, med, high to subdivide
	std::vector< std::vector<double> > corner_outputs(total_num_corners);
	//for each corner, need to deal with the outputs
	for(int i = 0; i < total_num_corners; i++)
		corner_outputs[i].resize(num_evaluators);

	//number of corners of the whole set of inputs, excluding the corners that come from regions out of subdivision
	int num_corners = (1 << input_dimension);	//2^n corners in the input space.  same number of regions (e.g. quadrants)
	std::vector<double> positive_corner_count(num_evaluators);	//used to count the number of positive values in a region

	//put midpoint in between low and high inputs
	std::vector<double> mid_input(input_dimension);
	for(int i = 0; i < input_dimension; i++)
		mid_input[i] = (max_input[i] + min_input[i]) / 2;

	//see if the subregion is within the tolerance for all inputs
	bool sub_region_accurate_enough = true;
	for(int i = 0; i < input_dimension; i++)
		if( fabs(max_input[i] - min_input[i]) > tolerance_input[i])
		{
			sub_region_accurate_enough = false;
			break;
		}

	//need to evaluate all functions at all corners
	std::vector<double> cur_input(input_dimension);	//used for constructing temporary input to compute corners
	for(int i = 0; i < total_num_corners; i++)
	{
		//figure out which vertex input to use
		for(int j = 0, d = 1; j < input_dimension; j++, d *= 3)
		{
			//pick the proper vertex for the corner using each digit in base 3 for low, med, high
			switch( (i/d) % 3)
			{
			case 0:		cur_input[j] = min_input[j];	break;
			case 1:		cur_input[j] = mid_input[j];	break;
			case 2:		cur_input[j] = max_input[j];	break;
			}
		}

		//evaluate the functions
		for(int j = 0; j < num_evaluators; j++)
			corner_outputs[i][j] = sre[j]->Evaluate(cur_input);
	}

	//go through every region
	bool found_valid_region = false;	//haven't found a valid region yet
	for(int i = num_corners-1; i >= 0; i--)	//prefer high values to low values (go through opposite way to prefer low values)
	{
		//first clear out all corner counts
		for(int j = 0; j < num_evaluators; j++)
			positive_corner_count[j] = 0;

		//take this specific combo of adding 1's to the digits in base 3
		int region_offset = 0;
		for(int k = 0; k < input_dimension; k++)
			if(i & (1 << k))
				region_offset += nth_pow_of_3(k);

		//count the number of corners that every eval function has a positive
		for(int j = 0; j < num_corners; j++)
		{
			//take the combo of adding digits to all the powers of 3
			int corner_offset = 0;
			for(int k = 0; k < input_dimension; k++)
				if(j & (1 << k))
					corner_offset += nth_pow_of_3(k);

			//check all evaluations in the corner for being positive
			int corner = region_offset + corner_offset;
			for(int m = 0; m < num_evaluators; m++)
				if(corner_outputs[corner][m] >= 0)
					positive_corner_count[m]++;
		}

		//take the the region if all eval functions have at least one sign not like the others
		bool use_this_region = true;
		for(int j = 0; j < num_evaluators; j++)
			if(positive_corner_count[j] == 0 || positive_corner_count[j] == num_corners)
			{
				use_this_region = false;
				break;
			}
		//if the region contains all sign changes or we haven't reached the minimum descent level, check inside
		if(use_this_region || minimum_descent_level > 0)
		{
			//make the function's own copies so it doesn't mess up those passed in
			std::vector<double> low_input(min_input);
			std::vector<double> high_input(max_input);

			//set low_input, high_input using results
			for(int j = 0, d = 1; j < input_dimension; j++, d *= 3)
			{
				//check to see if the region chosen has an offset for each dimension
				// and move the corresponding endpoint to the middle
				switch( (region_offset/d) % 3)
				{
				case 0:		high_input[j] = mid_input[j];	break;
				case 1:		low_input[j] = mid_input[j];	break;
				}
			}

			///if within range and using this sub-region, quit
			if(use_this_region && sub_region_accurate_enough)
			{
				std::copy(mid_input.begin(), mid_input.end(), results.begin());
				return true;
			}

			//recurse
			if(SolveSimultaneousRoots(sre, low_input, high_input, tolerance_input, results, minimum_descent_level - 1))
				return true;
		}
	} //for every region

	//didn't find any
	return false;
}

//takes in an Evaluator to use as functions to evaluate the inputs,
// as well as the vectors min_input and max_input for the bounds on each variable (each element position is a different variable)
// returns the result that maximizes the Evaluator, as found by a sort of continuous beam search
// The parameter minimum_descent_level may be used to force the algorithm to subdivide each space a minimum number of times
// to help find the global maximum - larger values are better for functions with many local maxima
double MaximizeEvaluator(Evaluator &eval,
							double min_input, double max_input,
							int num_initial_evaluations = 3,
							double tollerance = 0.001)
{
	while(max_input - min_input > tollerance)
	{
		//evaluate the function at each of the evaluations
		double max_output = -INFINITY;
		int max_offset = -1;
		std::vector<double> input(1);
		for(int i = 0; i < num_initial_evaluations; i++)
		{
			input[0] = min_input + (max_input - min_input) * (i + .5) / num_initial_evaluations;
			double output = eval.Evaluate(input);
			
			if(output > max_output)
			{
				max_output = output;
				max_offset = i;
			}
		}
		
		//now have the best input, squeeze the boundary around it
		min_input = std::max(min_input, min_input + (max_input - min_input) * (max_offset - 1 - .5) / num_initial_evaluations);
		max_input = std::min(max_input, min_input + (max_input - min_input) * (max_offset + 1 + .5) / num_initial_evaluations);
	}

	//return the average of the two best values
	return (min_input + max_input)/2;
}

#ifdef TEST_SOLVER

class TestEvaluator1 : public Evaluator
{
	double Evaluate(std::vector<double> v)
	{
		double x = v[0];
		double y = v[1];
		return y - 3 * x;
	}
};

class TestEvaluator2 : public Evaluator
{
	double Evaluate(std::vector<double> v)
	{
		double x = v[0];
		double y = v[1];
		return 0.25*y + x - 1 + 0.5*x*y;
	}
};

class TestEvaluator3 : public Evaluator
{
	double Evaluate(std::vector<double> v)
	{
		double x = v[0];
		double y = v[1];
		double z = v[2];
		return 0.25*z + x - 1 + 0.5*x*z;
	}
};

//answer: x = .4201, y = 1.260

void TestSolver()
{
	TestEvaluator1 t1;
	TestEvaluator2 t2;
	TestEvaluator3 t3;
	std::vector<Evaluator *> tests;
	tests.push_back(&t1);
	tests.push_back(&t2);

	std::vector<double> min;		min.push_back(0);			min.push_back(0);
	std::vector<double> max;		max.push_back(2);			max.push_back(2);
	std::vector<double> tolerance;	tolerance.push_back(0.00001);	tolerance.push_back(0.00001);
	std::vector<double> results;	results.push_back(0);		results.push_back(0);

	tests.push_back(&t3);
	min.push_back(0);
	max.push_back(2);
	tolerance.push_back(.001);
	results.push_back(0);

	if(SolveSimultaneousRoots(tests, min, max, tolerance, results, 1))
		for(int i = 0; i < results.size(); i++)
			printf("%f ", results[i]);
	else
		printf("no solution");
}

#endif //defined TEST_SOLVER
